#include <stdio.h>
#include <stdlib.h>

/* ADPCM decoding is under the following conditions */

/***********************************************************
Copyright 1992 by Stichting Mathematisch Centrum, Amsterdam, The
Netherlands.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the names of Stichting Mathematisch
Centrum or CWI not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior permission.

STICHTING MATHEMATISCH CENTRUM DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL STICHTING MATHEMATISCH CENTRUM BE LIABLE
FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

******************************************************************/

/* Alaw decoding is under the following conditions */
/*
** Copyright (C) 1999-2000 Erik de Castro Lopo <erikd@zip.com.au>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU Lesser General Public License as published by
** the Free Software Foundation; either version 2.1 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Lesser General Public License for more details.
**
** You should have received a copy of the GNU Lesser General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/**********************
record file layout
uint 0x10000037
uint uid2 0x100006d
uint uid3 0x1000007e
uint uid4 (checksum) 0x5508accf
uint offset to table section 0x00000014
uchar 0x04 (no uints in table)
uint 0x10000052 record entry
uint offset to record data (0x00000034)
uint 0x1000089 APP id section
uint offset to app id section (0x00000025)

unit 1000007e record
uchars  2a 52 65 63 6f 72 64 2e 61 70 70 ("*Record.app")

Can write above as one chunk of 52 chars

uint uncompressed length
uint 0x00000000 alaw, 0x100001a1 adpcm
ushort repetitions - 1 (0x0)
uchar vol 0x01-05
uchar 0x00
uint millionths of sec between repeats
uint number of bytes following

ADPCM format (IMA standard) - in frames

cardinals decompressed_length, actual length
uint samples_in_frame

codec reset to zero after each frame


NB shorts & inst in rverse order in file
e.g. 0x00000001 => 01 00 00 00
*******************************/

typedef struct
{
    unsigned int frmsamples,cursample,nsamples,delay;
    unsigned int (*dec)(unsigned int, short*, void*);
    FILE *ifp;
    short valprev;
    unsigned short reps;
    unsigned char vol;
    char index;
} SF_STATE;

/* Copied from LIBSNDFILE */
static short alaw2pcm[256] =
{
     5504,  5248,  6016,  5760,  4480,  4224,  4992,  4736,
     7552,  7296,  8064,  7808,  6528,  6272,  7040,  6784,
     2752,  2624,  3008,  2880,  2240,  2112,  2496,  2368,
     3776,  3648,  4032,  3904,  3264,  3136,  3520,  3392,
    22016, 20992, 24064, 23040, 17920, 16896, 19968, 18944,
    30208, 29184, 32256, 31232, 26112, 25088, 28160, 27136,
    11008, 10496, 12032, 11520,  8960,  8448,  9984,  9472,
    15104, 14592, 16128, 15616, 13056, 12544, 14080, 13568,
      344,   328,   376,   360,   280,   264,   312,   296,
      472,   456,   504,   488,   408,   392,   440,   424,
       88,    72,   120,   104,    24,     8,    56,    40,
      216,   200,   248,   232,   152,   136,   184,   168,
     1376,  1312,  1504,  1440,  1120,  1056,  1248,  1184,
     1888,  1824,  2016,  1952,  1632,  1568,  1760,  1696,
      688,   656,   752,   720,   560,   528,   624,   592,
      944,   912,  1008,   976,   816,   784,   880,   848,
     -5504,  -5248,  -6016,  -5760,  -4480,  -4224,  -4992,  -4736,
     -7552,  -7296,  -8064,  -7808,  -6528,  -6272,  -7040,  -6784,
     -2752,  -2624,  -3008,  -2880,  -2240,  -2112,  -2496,  -2368,
     -3776,  -3648,  -4032,  -3904,  -3264,  -3136,  -3520,  -3392,
    -22016, -20992, -24064, -23040, -17920, -16896, -19968, -18944,
    -30208, -29184, -32256, -31232, -26112, -25088, -28160, -27136,
    -11008, -10496, -12032, -11520,  -8960,  -8448,  -9984,  -9472,
    -15104, -14592, -16128, -15616, -13056, -12544, -14080, -13568,
      -344,   -328,   -376,   -360,   -280,   -264,   -312,   -296,
      -472,   -456,   -504,   -488,   -408,   -392,   -440,   -424,
       -88,    -72,   -120,   -104,    -24,     -8,    -56,    -40,
      -216,   -200,   -248,   -232,   -152,   -136,   -184,   -168,
     -1376,  -1312,  -1504,  -1440,  -1120,  -1056,  -1248,  -1184,
     -1888,  -1824,  -2016,  -1952,  -1632,  -1568,  -1760,  -1696,
      -688,   -656,   -752,   -720,   -560,   -528,   -624,   -592,
      -944,   -912,  -1008,   -976,   -816,   -784,   -880,   -848
};

/* Copied from CWI */
static int indexTable[16] = {
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
};

static int stepsizeTable[89] = {
    7, 8, 9, 10, 11, 12, 13, 14, 16, 17,
    19, 21, 23, 25, 28, 31, 34, 37, 41, 45,
    50, 55, 60, 66, 73, 80, 88, 97, 107, 118,
    130, 143, 157, 173, 190, 209, 230, 253, 279, 307,
    337, 371, 408, 449, 494, 544, 598, 658, 724, 796,
    876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066,
    2272, 2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358,
    5894, 6484, 7132, 7845, 8630, 9493, 10442, 11487, 12635, 13899,
    15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767
};

unsigned int ReadCardinal(FILE *ifp)
{
    unsigned int a;
    unsigned char byte;

    fread(&byte,1,1,ifp);
    a = byte;
    if(!(a&1))
        a>>=1;
    else
    {
        fread(&byte,1,1,ifp);
        a|= byte<<8;
        if(!(a&2))
            a>>=2;
        else if(!(a&4))
        {
            fread(&byte,1,1,ifp);
            a|=byte<<16;
            fread(&byte,1,1,ifp);
            a|=byte<<24;
            a>>=3;
        }
    }
    return a;
}

unsigned int ALAWdec(unsigned int len,short *obuff, void *v)
{
    SF_STATE *st = (SF_STATE *)v;
    unsigned int i=0;
    unsigned char byte;
    while(i<len && fread(&byte,1,1,st->ifp)==1)
    {
        *obuff++=alaw2pcm[byte];
        i++;
    }
    return i;
}

unsigned int ADPCMdec(unsigned int len,short *obuff, void *v)
{
    SF_STATE *st = (SF_STATE *)v;
    unsigned int i=0,ok=1;
    int sign;            /* Current adpcm sign bit */
    int delta;            /* Current adpcm output value */
    int step;            /* Stepsize */
    int valpred;        /* Predicted value */
    int vpdiff;            /* Current change to valpred */
    int index;            /* Current step change index */
    unsigned char inputbuffer;        /* place to keep next 4-bit value */
    int bufferstep;        /* toggle between inputbuffer/input */

    while(i<len && ok)
    {
        if(st->frmsamples == st->cursample)
        {
            st->frmsamples = ReadCardinal(st->ifp);
            ReadCardinal(st->ifp);
            fseek(st->ifp,4,SEEK_CUR);
            st->cursample = 0;
            st->valprev=0;st->index=0;
        }
        /* decode ADPCM */
        valpred = st->valprev;
        index = st->index;
        step = stepsizeTable[index];
        bufferstep = 0;
        while(i<len && st->cursample<st->frmsamples)
        {
            /* Step 1 - get the delta value */
            if ( bufferstep )
                delta = inputbuffer & 0xf;
            else
            {
                if(fread(&inputbuffer,1,1,st->ifp) != 1)
                    {ok=0;break;}
                delta = (inputbuffer >> 4) & 0xf;
            }
            bufferstep = !bufferstep;
            /* Step 2 - Find new index value (for later) */
            index += indexTable[delta];
            if ( index < 0 )
                index = 0;
            if ( index > 88 )
                index = 88;

            /* Step 3 - Separate sign and magnitude */
            sign = delta & 8;
            delta = delta & 7;

            /* Step 4 - Compute difference and new predicted value */
            /*
            ** Computes 'vpdiff = (delta+0.5)*step/4', but see comment
            ** in adpcm_coder.
            */
            vpdiff = step >> 3;
            if ( delta & 4 )
                vpdiff += step;
            if ( delta & 2 )
                vpdiff += step>>1;
            if ( delta & 1 )
                vpdiff += step>>2;

            if ( sign )
                valpred -= vpdiff;
            else
                valpred += vpdiff;

            /* Step 5 - clamp output value */
            if ( valpred > 32767 )
                valpred = 32767;
            else if ( valpred < -32768 )
                valpred = -32768;

            /* Step 6 - Update step value */
            step = stepsizeTable[index];

            /* Step 7 - Output value */
            *obuff++ = valpred;
            st->cursample++;
            i++;
        }
    }
    return i;
}

int OpenSoundFile(FILE *ifp, SF_STATE *st)
{
    int i,j,len = -1,offset;
    unsigned int rdi;
    unsigned short rds;
    unsigned char rdc;

/* UIDs */
    fread(&rdi,4,1,ifp);
    if(rdi != 0x10000037)
        return len;
    fread(&rdi,4,1,ifp);
    if(rdi != 0x1000006d)
        return len;
    fread(&rdi,4,1,ifp);
    if(rdi != 0x1000007e)
        return len;
    fread(&rdi,4,1,ifp);
    if(rdi != 0x5508accf)
        return len;
/* table section */
    fread(&rdi,4,1,ifp);
    fseek(ifp,rdi,SEEK_SET);
    fread(&rdc,1,1,ifp);
    for(i=rdc/2;i>0;i--)
    {
        fread(&rdi,4,1,ifp);
        switch (rdi)
        {
            case 0x10000052: /* record Data offset */
                fread(&offset,4,1,ifp);
                break;
            case 0x10000089: /* APP ID */
                fread(&rdi,4,1,ifp);
                j=ftell(ifp);
                fseek(ifp,rdi,SEEK_SET);
                fread(&rdi,4,1,ifp);
                if(rdi != 0x1000007e) /* recorded by record */
                    return len;
                fseek(ifp,j,SEEK_SET);
                break;
            default:
                return len;
                break;
        }
    }
    fseek(ifp,offset,SEEK_SET);
/* uncompressed length */
    fread(&rdi,4,1,ifp);
    len = rdi;
    st->nsamples = rdi;
/* compression */
    fread(&rdi,4,1,ifp);
    if(rdi == 0x00000000) /* Alaw */
    {st->dec=ALAWdec;}
    else if(rdi == 0x100001a1) /* ADPCM */
    {st->dec = ADPCMdec;st->cursample=0;st->frmsamples=0;}
    else /* unknown */
        return -1;
    fread(&rds,2,1,ifp);
    st->reps = rds;
    fread(&rdc,1,1,ifp);
    st->vol = rdc;
    fseek(ifp,1,SEEK_CUR);
    fread(&rdi,4,1,ifp);
    st->delay = rdi;
    fseek(ifp,4,SEEK_CUR);
    st->ifp = ifp;
    return len;
}

void WriteWAVHeader(FILE *ofp, SF_STATE *st)
{
    short s;
    int i;
    fwrite("RIFF",1,4,ofp);
    i=st->nsamples*2 + 36;
    fwrite(&i,4,1,ofp);//following length
    fwrite("WAVE",1,4,ofp);
    fwrite("fmt ",1,4,ofp);
    i=16;
    fwrite(&i,4,1,ofp);//size of fmt block
    s=1;
    fwrite(&s,2,1,ofp);//format
    s=1;
    fwrite(&s,2,1,ofp);//channels
    i=8000;
    fwrite(&i,4,1,ofp);//sampling rate
    i=16000;
    fwrite(&i,4,1,ofp);//avg bytes per sec
    s=2;
    fwrite(&s,2,1,ofp);//block align
    s=16;
    fwrite(&s,2,1,ofp);//bits per sample
    fwrite("data",1,4,ofp);
    i=st->nsamples*2;
    fwrite(&i,4,1,ofp);
    return;
}

int main(int argc, char **argv)
{
    int i;
    FILE *ifp,*ofp;
    SF_STATE st;
    short *buffer;

    if(argc !=3)
        {printf("%s\n\nConverts Psion 5 Record files to WAV files.\n\n" \
                "USAGE %s infile outfile\n" \
                "Comments to Danny <drsmith@iee.org>\n",
                argv[0],argv[0]);return 0;}
    ifp=fopen(argv[argc-2],"rb");
    if(ifp == NULL)
        {printf("Error: could not open input file %s\n",argv[argc-2]);return -1;}
    i=OpenSoundFile(ifp,&st);
    if(i==-1)
        {printf("Error: input file is not correct format\n");fclose(ifp);return -1;}

    ofp=fopen(argv[argc-1],"wb");
    if(ofp == NULL)
        {printf("Error: could not open output file %s\n",argv[argc-1]);fclose(ifp);return -1;}
    buffer = malloc(4000*sizeof(short));
    if(buffer == NULL)
        {printf("Error: could not allocate buffer\n");fclose(ifp);fclose(ofp);return -1;}
    WriteWAVHeader(ofp,&st);
    do
    {
        i=st.dec(4000,buffer,&st);
        fwrite(buffer,2,i,ofp);
    } while(i == 4000);
    fclose(ifp);
    fclose(ofp);
    free(buffer);
    return 0;
}

